import logging
import datetime
import pytz
import dateutil
from dateutil.parser import parse as dateparser
import re

logger = logging.getLogger(__name__)


# TODO: Same date in all examples
# TODO: Move examples to separate documentation page?
# TODO: Move Zulustring into main class


class ZuluError(Exception):
    pass


class Zulu(datetime.datetime):
    # TODO: Revise documentation
    """Timezone aware datetime objects in UTC

    From constructor::

        Zulu() or Zulu.now()
            Zulu(2020, 5, 21, 20, 5, 31, 930343, tzinfo=<UTC>)

        Zulu(2020, 5, 12)
            Zulu(2020, 5, 12, 0, 0, tzinfo=<UTC>)

        Zulu(2020, 5, 21, 20, 5, 31)
            Zulu(2020, 5, 21, 20, 5, 31, tzinfo=<UTC>)

        Zulu('20200521T202041u590718Z')
            Zulu(2020, 5, 21, 20, 20, 41, 590718, tzinfo=<UTC>)

    Zulu String is on the format ``<date>T<time>u<microseconds>Z``, and is
    \'designed\'
    to be file name and double click friendly, as well as easily recognizable
    within some string when using regular expressions.
    Printing a Zulu object will result in this string, and the constructor
    can also be instantiated with it::

        z = Zulu(2020, 5, 12)
        print(z)
            20200512T000000u000000Z

        Zulu('20200521T202041u590718Z')
            Zulu(2020, 5, 21, 20, 20, 41, 590718, tzinfo=<UTC>)

    For an ISO 8601 formatted string, use custom function::

        z = Zulu('20200521T202041u590718Z')
        z.iso()
            '2020-05-21T20:20:41.590718+00:00'

    ISO strings can be parsed with a dedicated function, or with constructor::

        z = Zulu.parse_iso('2020-05-21T20:20:41.590718+00:00')
        print(z)
            20200521T202041u590718Z

        z = Zulu('2020-05-21T20:20:41.590718+00:00')
        print(z)
            20200521T202041u590718Z


    Inputs or constructors may vary, but Zulu objects are always UTC. Hence
    the name Zulu.

    Constructor also takes regular datetime objects, provided they have
    timezone info::

        dt = datetime.datetime(2020, 5, 23, tzinfo=pytz.utc)
        Zulu(dt)
            Zulu(2020, 5, 23, 0, 0, tzinfo=<UTC>)

        dt = datetime.datetime(2020, 5, 23, tzinfo=dateutil.tz.tzlocal())
        Zulu(dt)
            Zulu(2020, 5, 22, 22, 0, tzinfo=<UTC>)

    Examples::

        z = Zulu()
        print(z)
            20200522T190137u055918Z
        z.month
            5
        z.str.month
            '05'
        z.str.date
            '20200522'
        z.add(hours=2)
            Zulu(2020, 5, 22, 21, 1, 37, 55918, tzinfo=<UTC>)

    ``Zulu.elf()`` takes the same inputs as constructor, but also allowes
    Zulu objects to pass through. If timezone is missing, it will assume
    local. It can also be told to assume a specific timezone.

    """

    ZTR_FORMAT_STRING = '%Y%m%dT%H%M%Su%fZ'

    ZTR_REGEX_STRING = r'\d{8}T\d{6}u\d{6}Z'
    ZTR_REGEX = re.compile(ZTR_REGEX_STRING)

    ISO_REGEX_STRING = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-' \
                       r'(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):' \
                       r'([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-]' \
                       r'(?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
    ISO_REGEX = re.compile(ISO_REGEX_STRING)

    class ZuluString:
        """Initializes convenience strings, and is instantiated with each Zulu
        object and stored in ``Zulu.str``. Meant as the string version of the
        standard datetime.datetime integer values.

        Examples::

            z = Zulu('20200522T190137u055918Z')
            z.year
                2020
            z.str.year
                '2020'
            z.month
                5
            z.str.month
                '05'
            z.day
                22
            z.str.day
                '22'
            z.date()
                datetime.date(2020, 5, 22)
            z.str.date
                '20200522'
            z.time()
                datetime.time(19, 1, 37, 55918)
            z.str.time
                '190137'
            z.str.microsecond
                '055918'

            isinstance(z.str, ZuluString)
                True

        """

        def __init__(self, zulu):
            # Year
            self.year = f'{zulu.year:04d}'
            self.month = f'{zulu.month:02d}'
            self.day = f'{zulu.day:02d}'
            self.hour = f'{zulu.hour:02d}'
            self.minute = f'{zulu.minute:02d}'
            self.second = f'{zulu.second:02d}'
            self.microsecond = f'{zulu.microsecond:06d}'
            self.date = self.year + self.month + self.day
            self.time = self.hour + self.minute + self.second

            self.stub = str(zulu)

            self._zulu_repr = zulu.__repr__()

        def __str__(self):
            return self.stub

        def __repr__(self):
            return f'ZuluString({self._zulu_repr})'

    ############################################################################
    # String methods
    ############################################################################

    @classmethod
    def is_stub(cls,
                ts: str):
        """ Checks if string is Zulu string

        .. note::  The function looks for Zulu String, not ISO 8601

        :param ts: String that might be Zulu
        :type ts: str
        :return: True if string is Zulu, False if not
        :rtype: bool
        """
        return len(ts) == 23 and cls.ZTR_REGEX.match(ts) is not None

    @classmethod
    def is_in(cls,
              text: str):
        """ Check if input text contains a zulu stub

        .. note::  The function looks for Zulu String, not ISO 8601

        :type text: str
        :return: bool
        """
        if cls.ZTR_REGEX.search(text):
            return True
        else:
            return False

    @classmethod
    def find_one(cls,
                 text: str):
        """ Looks for and returns exactly one Zulu from text

        Will fail if there are none or multiple zulu\'s.
        Use find_all() to return a list of zulu\'s in text, including
        an empty list if there are none.

        .. note::  The function looks for Zulu String, not ISO 8601

        :type text: str
        :return: Zulu
        """
        res = cls.find_all(text)
        if not res:
            raise ZuluError(f'No zulu found in this text: {text}. '
                            f'Consider using find_all, which will return '
                            f'empty list if no zulu\'s are found.')
        elif len(res) == 1:
            return res[0]
        else:
            n = len(res)
            raise ZuluError(f'Found more than one (i.e. {n}) '
                            f'zulu in text: {text}. '
                            f'Consider using find_all, which will return '
                            f'a list of all zulu\'s found.')


    @classmethod
    def find_all(cls,
                 text: str):
        """ Finds and returns all zulu\'s in text

        Returns empty list if none are found.

        .. note::  The function looks for Zulu String, not ISO 8601

        :type text: str
        :return: [Zulu]
        """
        ztrs = re.findall(cls.ZTR_REGEX_STRING, text)
        return [cls.from_stub(x) for x in ztrs]

    @classmethod
    def is_iso(cls, ts_str):
        # TODO: Use regex instead, and include check on timezone (optional)
        """Tests if the input string is in fact an
        `ISO 8601 <https://en.wikipedia.org/wiki/ISO_8601>`_ string

        # .. note:: The test is made by converting the string to datetime
        #     and back again, making sure they are equal

        :param ts_str: Maybe an ISO formatted string
        :return: bool
        """
        try:
            if cls.ISO_REGEX.match(ts_str) is not None:
                return True
        except:
            pass
        return False
        # try:
        #     ts = Zulu._parse_iso(ts_str)
        #     return ts.isoformat() == ts_str
        # except ValueError:
        #     return False

    ############################################################################
    # Timezone methods
    ############################################################################

    @classmethod
    def all_timezones(cls):
        """ Returns a list of all allowed timezone names, except \'local\',
        which will return a datetime object with local timezone

        Wrapper for ``pytz.all_timezones``

        :return: list
        """
        return pytz.all_timezones

    @classmethod
    def _to_utc(cls, ts):
        return ts.astimezone(pytz.utc)

    @classmethod
    def _tz_from_name(cls, tz='utc'):
        if tz == 'local':
            tz = dateutil.tz.tzlocal()
        else:
            try:
                tz = pytz.timezone(tz)
            except pytz.exceptions.UnknownTimeZoneError as ue:
                raise ZuluError(f'Unknown timezone: \'{tz}\'. '
                                f'Use Zulu.all_timezones() for a list '
                                f'of actual timezones')
        return tz

    ############################################################################
    # Create methods
    ############################################################################

    @classmethod
    def now(cls,
            tz=None):
        """
        Overrides ``datetime.datetime.now()``. Equivalent to ``Zulu()``

        :raise ZuluError: If parameter ``tz`` has a value
        :param tz: Do not use. Zulu is always UTC
        :return: Zulu
        """
        if tz:
            raise ZuluError(f'Zulu.now() does not allow input time zone info. '
                            f'Zulu is always UTC. Hence the name')
        return cls()

    @classmethod
    def _from_unaware(cls, ts, tz=None):
        if not tz:
            raise ZuluError('No timezone info. Set timezone to use '
                            'with \'tz=<timezone string>\'. \'local\' will '
                            'use local timezone info. Use '
                            'Zulu.all_timezones() for a list of actual '
                            'timezones')
        return ts.replace(tzinfo=cls._tz_from_name(tz))

    @classmethod
    def _elf(cls, ts, tz=None):
        # Takes a datetime.datetime object and adds the input tzinfo if
        # none is present
        if not ts.tzinfo:
            ts = cls._from_unaware(ts, tz=tz)
        return ts

    @classmethod
    def from_unaware(cls, ts, tz='utc'):
        """ Create Zulu from timezone unaware datetime

        :param ts: Unaware time stamp
        :type ts: datetime.datetime
        :param tz: Time zone, with 'utc' as default.
            'local' will use local time zone
        :return: Zulu
        """
        if ts.tzinfo:
            raise ZuluError(f'Input datetime already has '
                            f'time zone info: {ts}. '
                            f'Use constructor or Zulu.elf()')
        else:
            ts = cls._from_unaware(ts, tz=tz)
        return cls(ts)

    @classmethod
    def from_unaware_local(cls, ts):
        """ Create Zulu from timezone unaware local timestamp

        :param ts: Timezone unaware datetime
        :type ts: datetime.datetime
        :return: Zulu
        """
        return cls.from_unaware(ts, tz='local')

    @classmethod
    def from_unaware_utc(cls, ts):
        """ Create Zulu from timezone unaware UTC timestamp

        :param ts: Timezone unaware datetime
        :type ts: datetime.datetime
        :return: Zulu
        """
        return cls.from_unaware(ts, tz='utc')

    @classmethod
    def _parse_iso(cls,
                   iso: str):
        ts = dateparser(iso)
        if ts.tzinfo and str(ts.tzinfo) == 'tzutc()':
            ts = ts.astimezone(pytz.utc)
        return ts

    @classmethod
    def from_iso(cls,
                 iso: str,
                 tz=None):
        """

        :param iso: ISO 8601 string
        :param tz: Timezone string to use if missing in ts_str
        :return:
        """
        ts = cls._parse_iso(iso)
        if tz and not ts.tzinfo:
            ts = cls._from_unaware(ts, tz)
        elif ts.tzinfo and tz:
            raise ZuluError(f'Timezone info found in ISO string as well as '
                            f'input timezone argument (tz). Keep tz=None, '
                            f'or use Zulu.elf()')
        elif not tz and not ts.tzinfo:
            raise ZuluError('No timezone info in neither ISO string '
                            'nor tz argument')
        return cls(ts)

    @classmethod
    def _parse(cls,
               ts_str: str,
               pattern: str):
        return datetime.datetime.strptime(ts_str, pattern)

    @classmethod
    def parse(cls,
              ts_str: str,
              pattern: str,
              tz=None):
        """Parse time stamp string with the given pattern

        :param ts_str: Timestamp string
        :type ts_str: str
        :param pattern: Follows standard
            `python strftime reference <https://strftime.org/>`_
        :param tz: Timezone to use if timestamp does not have timezone info
        :return: Zulu
        """
        ts = cls._parse(ts_str, pattern)
        if not ts.tzinfo:
            ts = cls._from_unaware(ts, tz=tz)
        elif tz:
            raise ZuluError('Cannot have an input timezone argument when '
                            'input string already has timezone information')
        return cls(ts)

    @classmethod
    def _from_stub(cls, ztr):
        ts = cls._parse(ztr, cls.ZTR_FORMAT_STRING)
        return cls._from_unaware(ts, tz='utc')

    @classmethod
    def from_stub(cls,
                  ztr: str):
        if not cls.is_stub(ztr):
            raise ZuluError(f'String is not ztr: {ztr}')
        ts = cls._from_stub(ztr)
        return cls(ts)

    @classmethod
    def _from_epoch(cls, epoch):
        ts = datetime.datetime.utcfromtimestamp(epoch)
        return ts.replace(tzinfo=pytz.UTC)

    @classmethod
    def from_epoch(cls, epoch):
        ts = cls._from_epoch(epoch)
        return cls(ts)

    @classmethod
    def _fill_args(cls, args):
        if len(args) < 8:
            # From date
            args = list(args)
            args += (8-len(args))*[0]
            if args[1] == 0:
                args[1] = 1
            if args[2] == 0:
                args[2] = 1
            args = tuple(args)

        if args[-1] not in [None, 0, pytz.utc]:
            raise ZuluError(f'Zulu can only be UTC. '
                            f'Invalid timezone: {args[-1]}')

        args = list(args)
        args[-1] = pytz.utc
        return tuple(args)

    @classmethod
    def elf(cls, *args, tz='local'):
        """General input Zulu constructor

        .. warning:: Elves are fickle

        Handles datetime.datetime, Zulu string, ISO 8601 string and Unix Epoch
        the same way as constructor. However, elf will assume the default
        timezone given in argument ``tz`` if the input has no timezone info.
        Default is ``tz='local'``

        :param args: Input arguments
        :param tz: Time zone to assume if missing. 'local' will use local
            time zone. Use Zulu.all_timezones() for a list of actual
            timezones. Default is 'local'
        :return: Zulu
        """
        if len(args) == 0:
            return cls()
        elif len(args) > 1:
            args = cls._fill_args(args)
            ts = datetime.datetime(*args)
            if not ts.tzinfo:
                ts = cls._from_unaware(ts, tz)
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, Zulu):
                return arg
            elif isinstance(arg, datetime.datetime):
                # Add timzone if missing
                ts = cls._elf(arg, tz=tz)
                return cls(ts)
            elif isinstance(arg, float):
                return cls.from_epoch(arg)
            elif isinstance(arg, int):
                # Instantiate as start of year
                return cls(arg, 1, 1)
            elif isinstance(arg, str):
                if cls.is_stub(arg):
                    return cls.from_stub(arg)
                elif cls.is_iso(arg):
                    ts = cls._parse_iso(arg)
                    # Add timzone if missing
                    ts = cls._elf(ts, tz=tz)
                else:
                    raise ZuluError(f'String is neither zulu, nor ISO: {arg}. '
                                    f'Use Zulu.parse(). Or a different string')
            else:
                raise ZuluError(f'Found no way to interpret input '
                                f'argument as Zulu: {arg} [{type(arg)}]')
        return cls(ts)


    @classmethod
    def range(cls,
              start=None,
              n=10,
              delta=datetime.timedelta(hours=1)):
        """Generate a list of Zulu of fixed intervals

        ..note:: Mainly for dev purposes. There are far better
            ways of creating a range of timestamps, such as using pandas.

        :param start: Start time Zulu, default is *now*
        :type start: Zulu
        :param n: Number of timestamps in range, with default 10
        :type n: int
        :param delta: Time delta between items, with default one hour
        :type delta: datetime.timedelta
        :return: [Zulu]
        """
        if not start:
            start = cls()
        return [Zulu.elf(start + x * delta) for x in range(n)]

    def __new__(cls, *args, **kwargs):
        # TODO: Cleanup instantiation (ZuluString is used three times)
        # TODO: Cleanup instantiation with only datetime and digits
        if len(args) == 0 and len(kwargs) == 0:
            ts = datetime.datetime.utcnow()
            ts = ts.replace(tzinfo=pytz.UTC)
        elif len(args) == 1 and len(kwargs) == 0:
            arg = args[0]
            if isinstance(arg, str):
                raise ZuluError('Cannot instantiate Zulu with a string. Use '
                                'Zulu.from_iso(), Zulu.from_stub(), '
                                'Zulu.from_string() or Zulu.parse()')
            elif isinstance(arg, float):
                raise ZuluError(f'Cannot create Zulu object from a float: '
                                f'{arg}; If float is unix epoch, '
                                f'use Zulu.epoch()')
            elif isinstance(arg, Zulu):
                raise ZuluError(f'Input argument is already Zulu: {arg}')
            elif isinstance(arg, datetime.datetime):
                ts = arg
                if not ts.tzinfo:
                    raise ZuluError('Cannot create Zulu from datetime if '
                                    'datetime object does not have timezone '
                                    'info. Use Zulu.from_unaware()')
                ts = ts.astimezone(pytz.UTC)
            else:
                raise ZuluError(f'Unable to interpret input argument: '
                                f'{arg} [{type(arg).__name__}]')
        else:
            # Handle input as regular datetime input (year, month, day etc)
            try:
                ts = datetime.datetime(*args)
            except TypeError as te:
                raise ZuluError from te
            # Add timezone info if missing (assume utc, of course)
            if not ts.tzinfo:
                ts = ts.replace(tzinfo=pytz.UTC)

        # Create actual object
        args = tuple([ts.year, ts.month, ts.day,
                      ts.hour, ts.minute, ts.second,
                      ts.microsecond, ts.tzinfo])
        self = super(Zulu, cls).__new__(cls, *args)
        self.str = cls.ZuluString(self)
        return self

    def __str__(self):
        return self.strftime(self.ZTR_FORMAT_STRING)

    def __repr__(self):
        times = [self.hour, self.minute, self.second]
        has_micro = self.microsecond > 0
        has_time = sum(times) > 0
        nums = [self.year, self.month, self.day]
        if has_time or has_micro:
            nums += times
        if has_micro:
            nums += [self.microsecond]
        numstr = ', '.join([str(x) for x in nums])
        return f'Zulu({numstr})'

    def epoch(self):
        """
        Get epoch (seconds since January 1st 1970)

        Wrapper for ``datetime.datetime.timestamp()``

        :return: Seconds since January 1st 1970
        :rtype: float
        """
        return self.timestamp()

    # TODO: Replace with stub in tests
    def stub(self):
        return str(self)

    @classmethod
    def from_string(cls, string):
        if cls.is_stub(string):
            return cls.from_stub(string)
        elif cls.is_iso(string):
            return cls.from_iso(string)
        else:
            raise ZuluError(f'Unknown string format (neither stub nor iso): '
                            f'{string}; '
                            f'Use Zulu.parse() to specify format pattern and '
                            f'timezone')

    def iso(self, full=False):
        # TODO: Implement full flag
        """Create `ISO 8601 <https://en.wikipedia.org/wiki/ISO_8601>`_ string

        Example::

            z = Zulu(2020, 5, 21)
            z.iso()
                '2020-05-21T00:00:00+00:00'

        :return: str
        """
        if full:
            raise ZuluError('Full isoformat not implemented. Full means it '
                            'has fixed length no matter the value')
        return self.isoformat()

    def format(self, pattern):
        """Format Zulu to string with the given pattern

        :param pattern: Follows standard
            `Python strftime reference <https://strftime.org/>`_
        :return: str
        """
        return self.strftime(pattern)

    def to_unaware(self):
        """
        Get timezone unaware datetime object in UTC

        :return: Timezone unaware datetime
        :rtype: datetime.datetime
        """
        return datetime.datetime.utcfromtimestamp(self.epoch())

    def to_tz(self, tz='local'):
        """ Create regular datetime with input timezone

        For a list of timezones use :meth:`.Zulu.all_timezones()`, which is a
        wrapper for

        :param tz: Time zone to use. 'local' will return the local time zone.
            Default is 'local'
        :return: datetime.datetime
        """
        ts_utc = datetime.datetime.utcfromtimestamp(self.epoch())
        ts_utc = ts_utc.replace(tzinfo=pytz.UTC)
        return ts_utc.astimezone(self._tz_from_name(tz))

    def to_local(self):
        """ Create regular datetime with local timezone

        :return: datetime.datetime
        """
        return self.to_tz(tz='local')

    @classmethod
    def delta(cls,
              days=0,
              hours=0,
              minutes=0,
              seconds=0,
              microseconds=0,
              weeks=0):
        """Wrapper for datetime.timedelta

        :param days: Number of days
        :param hours: Number of hours
        :param minutes: Number of minutes
        :param seconds: Number of seconds
        :param microseconds: Number of microseconds
        :param weeks: Number of weeks
        :return: datetime.timedelta
        """
        return datetime.timedelta(days=days,
                                  hours=hours,
                                  minutes=minutes,
                                  seconds=seconds,
                                  microseconds=microseconds,
                                  weeks=weeks)

    def add(self,
            days=0,
            hours=0,
            minutes=0,
            seconds=0,
            microseconds=0,
            weeks=0):
        """Adds the input to current Zulu object and returns a new one

        :param days: Number of days
        :param hours: Number of hours
        :param minutes: Number of minutes
        :param seconds: Number of seconds
        :param microseconds: Number of microseconds
        :param weeks: Number of weeks
        :return: Zulu
        """
        delta = self.delta(days=days,
                           hours=hours,
                           minutes=minutes,
                           seconds=seconds,
                           microseconds=microseconds,
                           weeks=weeks)
        return Zulu(self + delta)



import uuid
import re


class IdentityError(Exception):
    pass


class Identity(str):
    """ UUID string generator with convenience functions

    Inherits str, and is therefore an immutable string, with a fixed format
    as illustrated below.

    Examples::

        Identity()
            'BD8E446D_3EB9_4396_8173_FA1CF146203C'

        Identity.is_in('Has BD8E446D_3EB9_4396_8173_FA1CF146203C within')
            True

        Identity.find_one('Has BD8E446D_3EB9_4396_8173_FA1CF146203C within')
            'BD8E446D_3EB9_4396_8173_FA1CF146203C'

    """

    _REGEX_STRING = r'[0-9A-F]{8}\_[0-9A-F]{4}\_[0-9A-F]{4}\_[0-9A-F]{4}' \
                   r'\_[0-9A-F]{12}'
    _REGEX_STRING_EXACT = r'^' + _REGEX_STRING + r'$'
    _REGEX = re.compile(_REGEX_STRING)
    _REGEX_EXCACT = re.compile(_REGEX_STRING_EXACT)

    @classmethod
    def is_stub(cls,
                stub: str):
        """ Check if input text is an identity

        :type stub: str
        :return: bool
        """
        if cls._REGEX_EXCACT.match(stub):
            return True
        else:
            return False

    @classmethod
    def is_in(cls,
              text: str):
        """ Check if input text contains an identity

        :type text: str
        :return: bool
        """
        if cls._REGEX.search(text):
            return True
        else:
            return False

    @classmethod
    def find_one(cls,
                 text: str):
        """ Looks for and returns exactly one Identity from text

        Will fail if there are none or multiple identities.
        Use find_all() to return a list of identities in text, including
        an empty list if there are none.

        :type text: str
        :return: Identity
        """
        res = cls._REGEX.search(text)
        if res:
            return cls(res.group())
        else:
            raise IdentityError(f'No identity found in this text: {text}. '
                                f'Consider using find_all, which will return '
                                f'empty list if no identities are found.')

    @classmethod
    def find_all(cls,
                 text: str):
        """ Finds and returns all identities in text

        :type text: str
        :return: [Identity]
        """
        ids = re.findall(cls._REGEX_STRING, text)
        return [cls(x) for x in ids]

    @classmethod
    def _verify_string(cls, identity_string):
        if cls._REGEX.match(identity_string) is None:
            raise IdentityError(f'String is not a valid identity: '
                                f'{identity_string}')

    @classmethod
    def from_stub(cls,
                  stub: str):
        return cls(stub)

    def __new__(cls,
                identity: str = None):
        if not identity:
            identity = str(uuid.uuid4()).replace('-', '_').upper()
        else:
            cls._verify_string(identity)
        instance = super(Identity, cls).__new__(cls, identity)
        return instance

    def __repr__(self):
        return f'Identity(\'{self}\')'

    def stub(self):
        return self.__str__()

    @classmethod
    def elf(cls, identity):
        """ Returns new or existing identity

        None will create a new Identity, a string will be parsed as Identity,
        and an existing Identity will be returned as itself.

        :type identity: None, str or Identity
        :return: Identity
        """
        if isinstance(identity, Identity):
            return identity
        else:
            return cls(identity)




from mjooln.core.name import Name, NameError
from mjooln.core.dic import Dic, DicError
from mjooln.core.doc import Doc, JSON, YAML, DocError
from mjooln.core.crypt import Crypt, CryptError

from mjooln.atom.identity import Identity, IdentityError
from mjooln.atom.zulu import Zulu, ZuluError
from mjooln.atom.element import Element, ElementError
from mjooln.atom.key import Key, InvalidKey
from mjooln.atom.atom import Atom, AtomError

from mjooln.filesystem.path import Path, PathError
from mjooln.filesystem.archive import Archive, ArchiveError
from mjooln.filesystem.folder import Folder, FolderError
from mjooln.filesystem.file import File, FileName, FileError
from mjooln.filesystem.filesystem import FileSystem, FileSystemError

from mjooln.tree.root import Root, RootError, NotRootException, ErraticWakeSleep
from mjooln.tree.tree import Tree, TreeError, NotTreeException
from mjooln.tree.leaf import Leaf, LeafError
from mjooln.tree.ground import Ground, GroundProblem, NoGround



import json
import simplejson
import yaml as YAML

from mjooln.core.name import Name
from mjooln.core.dic import Dic, DicError


class DocError(DicError):
    pass


class JSON:
    """Dict to/from JSON string, with optional human readable"""

    @classmethod
    def dumps(cls, dic, human=True, sort_keys=False, indent=4 * ' '):
        """Convert from dict to JSON string

        :param dic: Input dictionary
        :type dic: dict
        :param human: Human readable flag
        :param sort_keys: Sort key flag (human readable only)
        :param indent: Indent to use (human readable only)
        """
        if human:
            return simplejson.dumps(dic, sort_keys=sort_keys, indent=indent)
        else:
            return json.dumps(dic)

    @classmethod
    def loads(cls, json_string):
        """ Parse JSON string to dictionary

        :param json_string: JSON string
        :return: dict
        """
        return simplejson.loads(json_string)


class Dic:
    """Enables child classes to mirror attributes and dictionaries

    Private variables start with underscore, and are ignored by default.

    .. note:: Meant for inheritance and not direct use
    """

    # TODO: Consider moving ignore_private to a private class attribute
    # TODO: Consider deprecation. Or removing force equal/addonlyexisting.
    # TODO: Replace add and dic/doc with to_dict, from_dict, to_yaml etc
    # TODO: Require equal keys, except private.
    # Maybe even require equal keys if adding. I.e. only to be used for
    # configuration, or serialization
    _PRIVATE_STARTSWITH = '_'

    @classmethod
    def from_dict(cls,
                  data: dict):
        doc = cls()
        doc.add(data)
        return doc

    @classmethod
    def default(cls):
        return cls()

    @classmethod
    def default_dict(cls):
        dic = cls.default()
        return dic.to_dict()

    def to_vars(self):
        return vars(self).copy()

    def __repr__(self):
        dic = self.to_vars()
        dicstr = []
        for key, item in dic.items():
            dicstr.append(f'{key}={item.__repr__()}')
        dicstr = ', '.join(dicstr)
        return f'{type(self).__name__}({dicstr})'

    def to_dict(self,
                ignore_private: bool = True):
        """ Return dictionary with a copy of attributes

        :param ignore_private: Ignore private attributes flag
        :return: dict
        """
        dic = self.to_vars()
        if ignore_private:
            pop_keys = [x for x in dic
                        if x.startswith(self._PRIVATE_STARTSWITH)]
            for key in pop_keys:
                dic.pop(key)
        for key, item in dic.items():
            if isinstance(item, Dic):
                dic[key] = item.to_dict(ignore_private=ignore_private)
        return dic

    # @classmethod
    # def _from_strings(cls, dic):
    #     for key, item in dic.items():
    #         if isinstance(item, str):
    #             dic[key] = cls._parse_if_iso(item)
    #         elif isinstance(item, dict):
    #             dic[key] = cls._from_strings(item)
    #             if Atom.check(dic[key]):
    #                 dic[key] = Atom(**dic[key])
    #         elif isinstance(item, Dic):
    #             dic[key] = cls._from_strings(item.to_dict())
    #     return dic
    #
    # @classmethod
    # def _to_strings(cls, dic):
    #     for key, item in dic.items():
    #         if isinstance(item, Zulu):
    #             dic[key] = item.iso()
    #         elif isinstance(item, Identity):
    #             dic[key] = str(item)
    #         elif isinstance(item, Key):
    #             dic[key] = str(item)
    #         elif isinstance(item, Atom):
    #             dic[key] = cls._to_strings(vars(item))
    #         elif isinstance(item, dict):
    #             dic[key] = cls._to_strings(item)
    #         elif isinstance(item, Dic):
    #             dic[key] = cls._to_strings(item.to_dict())
    #     return dic

    def _add_item(self, key, item, ignore_private=True):
        # Add item and ignore private items if ignore_private is set to True
        if not ignore_private or not key.startswith(self._PRIVATE_STARTSWITH):
            self.__setattr__(key, item)

    def _add_dic(self,
                 dic: dict,
                 ignore_private: bool = True):
        for key, item in dic.items():
            self._add_item(key, item, ignore_private=ignore_private)

    def add(self,
            dic: dict,
            ignore_private: bool = True):
        """ Add dictionary to class as attributes

        :param dic: Dictionary to add
        :param ignore_private: Ignore private attributes flag
        :return: None
        """
        self._add_dic(dic, ignore_private=ignore_private)

    def add_only_existing(self, dic, ignore_private=True):
        """ Add dictionary keys and items as attributes if they already exist
        as attributes

        :param dic: Dictionary to add
        :param ignore_private: Ignore private attributes flag
        :return: None
        """
        dic_to_add = {}
        for key in dic:
            if hasattr(self, key):
                dic_to_add[key] = dic[key]
        self._add_dic(dic_to_add, ignore_private=ignore_private)

    def force_equal(self, dic, ignore_private=True):
        """ Add all dictionary keys and items as attributes in object, and
        delete existing attributes that are not keys in the input dictionary

        :param dic: Dictionary to add
        :param ignore_private: Ignore private attributes flag
        :return: None
        """
        self._add_dic(dic, ignore_private=ignore_private)
        for key in self.to_dict(ignore_private=ignore_private):
            if key not in dic:
                self.__delattr__(key)

    def print(self,
              ignore_private=True,
              indent=4*' ',
              width=80,
              flatten=False,
              separator=Name.ELEMENT_SEPARATOR):
        # TODO: Rename dev_print to print (print goes to terminal inherently)
        """ Pretty print of attributes in terminal meant for development
        purposes
        """
        text = f'--{indent}[[ {type(self).__name__} ]]{indent}'
        text += (width-len(text)) * '-'
        print(text)
        if not flatten:
            dic = self.to_dict(ignore_private=ignore_private)
        else:
            dic = self.flatten(sep=separator)
        self._print(dic, level=0)
        print(width*'-')

    def _print(self, dic, level=0, indent=4 * ' '):
        for key, item in dic.items():
            if isinstance(item, dict):
                self._print(item, level=level + 1)
            else:
                print(level*indent + f'{key}: [{type(item).__name__}] {item} ')

    def print_flat(self,
                   ignore_private=True,
                   separator=Name.ELEMENT_SEPARATOR):
        self.print(ignore_private=ignore_private,
                   separator=separator, flatten=True)

    # TODO: Move to flag in to_dict etc., and unflatten in from_dict etc
    # TODO: Replace sep with Key sep.
    # TODO: Require var names not to have double underscores
    # TODO: Figure out how to handle __vars__, what is the difference with _vars
    def flatten(self, sep=Name.ELEMENT_SEPARATOR, ignore_private=True):
        """
        Flatten dictionary to top level only by combining keys with the
        given separator

        :param sep: Separator to use, default is double underscore (__)
        :type sep: str
        :param ignore_private: Flags whether to ignore private attributes,
            identified by starting with underscore
        :return: Flattened dictionary
        :rtype: dict
        """
        dic = self.to_dict(ignore_private=ignore_private)
        flat_dic = dict()
        flat_dic = self._flatten(flat_dic=flat_dic, parent_key='',
                                 dic=dic, sep=sep)
        return flat_dic

    def _flatten(self,
                 flat_dic, parent_key, dic, sep=Name.ELEMENT_SEPARATOR):
        for key, item in dic.items():
            if isinstance(item, dict):
                flat_dic = self._flatten(flat_dic=flat_dic, parent_key=key,
                                         dic=item, sep=sep)
            else:
                if sep in key:
                    raise DicError(f'Separator \'{sep}\' found in '
                                   f'key: {key}')
                if parent_key:
                    flat_dic[parent_key + sep + key] = item
                else:
                    flat_dic[key] = item
        return flat_dic


# TODO: Rewrite as serializer/deserializer
# TODO: In other words, doc will be dic with dates as strings.
# TODO: Atom can be builtin, otherwise it will be a bit tricky.
# TODO: SHould zulu/key/identity also be builtin?
class Doc(Dic):
    """ Enables child classes to mirror attributes, dictionaries and JSON
    strings

    Special cases:

    - Zulu objects will be converted to an ISO 8601 string before a dictionary
      is converted to JSON
    - ISO 8601 strings that are time zone aware with UTC, will be converted to
      Zulu objects after JSON document has been converted to a dictionary
    - Elements that are dictionaries with key names corresponding to Atom
      (key, zulu, identity), will be recognized and converted back to an Atom
      object after JSON document has been converted to a dictionary
    """

    @classmethod
    def from_doc(cls, doc: dict):
        """
        Override in child class to handle non serializable items

        :param doc: Dictionary with serializable items only
        :return: New Doc object instantiated with input dictionary
        :rtype: Doc
        """
        return cls.from_dict(doc)

    @classmethod
    def from_json(cls,
                  json_string: str):
        doc = JSON.loads(json_string=json_string)
        return cls.from_doc(doc)

    @classmethod
    def from_yaml(cls,
                  yaml_string: str):
        doc = YAML.safe_load(yaml_string)
        return cls.from_doc(doc)

    def to_dict(self,
                ignore_private: bool = True):
        """ Return dictionary with a copy of attributes

        :param ignore_private: Ignore private attributes flag
        :return: dict
        """
        doc = vars(self).copy()
        if ignore_private:
            pop_keys = [x for x in doc
                        if x.startswith(self._PRIVATE_STARTSWITH)]
            for key in pop_keys:
                doc.pop(key)
        return doc

    def to_doc(self,
               ignore_private: bool = True):
        doc = vars(self).copy()
        if ignore_private:
            pop_keys = [x for x in doc
                        if x.startswith(self._PRIVATE_STARTSWITH)]
            for key in pop_keys:
                doc.pop(key)
        for key, item in doc.items():
            if isinstance(item, Doc):
                doc[key] = item.to_doc(ignore_private=ignore_private)
        return doc

    def to_json(self,
                human_readable: bool = False,
                ignore_private: bool = True):
        doc = self.to_doc(ignore_private=ignore_private)
        return JSON.dumps(doc, human=human_readable)

    def to_yaml(self,
                ignore_private: bool = True):
        doc = self.to_doc(ignore_private=ignore_private)
        return YAML.safe_dump(doc)




class NameError(Exception):
    pass


class Name:
    MINIMUM_ELEMENT_LENGTH = 2
    ELEMENT_SEPARATOR = '__'
    CLASS_SEPARATOR = '___'


class Stub:

    @classmethod
    def from_stub(cls, stub: str):
        raise NameError(f'Method from_stub() must be overridden in child '
                        f'class \'{cls.__name__}\'')

    def stub(self):
        raise NameError(f'Method stub() must be overridden in child '
                        f'class \'{type(self).__name__}')




import logging
import string

from mjooln.core.name import Name

logger = logging.getLogger(__name__)


class ElementError(Exception):
    pass


class Element(str):
    """
    Defines element string with limitations

    - Minimum length is 3
    - Allowed characters are
        - Lower case ascii (a-z)
        - Digits (0-9)
        - Underscore (_)
    - Underscore and digits can not be the first character
    - Underscore can not be the last character
    - Can not contain double underscore since it acts as separator for elements
      in Key

    Sample elements::

        'simple'
        'with_longer_name'
        'digit1'
        'longer_digit2'

    """

    ALLOWED_CHARACTERS = string.ascii_lowercase + string.digits + '_'
    ALLOWED_STARTSWITH = string.ascii_lowercase
    ALLOWED_ENDSWITH = string.ascii_lowercase + string.digits

    def __new__(cls,
                element: str):
        # TODO: Add list as input, creating key with separator
        cls.verify_element(element)
        self = super(Element, cls).__new__(cls, element)
        return self

    def __repr__(self):
        return f'Element(\'{self}\')'

    def stub(self):
        return self.__str__()

    @classmethod
    def from_stub(cls, stub: str):
        return cls(stub)

    @classmethod
    def verify_element(cls, element: str):
        if not len(element) >= Name.MINIMUM_ELEMENT_LENGTH:
            raise ElementError(f'Element too short. Element \'{element}\' has '
                               f'length {len(element)}, while minimum length '
                               f'is {Name.MINIMUM_ELEMENT_LENGTH}')
        if not element[0] in cls.ALLOWED_STARTSWITH:
            raise ElementError(f'Invalid startswith. Element \'{element}\' '
                               f'cannot start with \'{element[0]}\'. '
                               f'Allowed startswith characters are: '
                               f'{cls.ALLOWED_STARTSWITH}')
        if not element[-1] in cls.ALLOWED_ENDSWITH:
            raise ElementError(f'Invalid endswith. Element \'{element}\' '
                               f'cannot end with \'{element[-1]}\'. '
                               f'Allowed endswith characters are: '
                               f'{cls.ALLOWED_ENDSWITH}')
        invalid_characters = [x for x in element if x not in
                              cls.ALLOWED_CHARACTERS]
        if len(invalid_characters) > 0:
            raise ElementError(
                f'Invalid character(s). Element \'{element}\' cannot '
                f'contain any of {invalid_characters}. '
                f'Allowed characters are: '
                f'{cls.ALLOWED_CHARACTERS}')
        if Name.ELEMENT_SEPARATOR in element:
            raise ElementError(f'Element contains element separator, which is '
                               f'reserved for separating Elements in a Key.'
                               f'Element \'{element}\' cannot contain '
                               f'\'{Name.CLASS_SEPARATOR}\'')

    @classmethod
    def is_stub(cls,
                element: str):
        try:
            cls.verify_element(element)
            return True
        except ElementError:
            return False

    @classmethod
    def elf(cls, element):
        """ Allows key class to pass through instead of throwing exception

        :param element: Input element string or element class
        :type element: str or Element
        :return: Element
        """
        # TODO: Handle input and guess a conversion that would match criteria
        if isinstance(element, Element):
            return element
        else:
            return cls(element)



import logging

from mjooln.core.name import Name
from mjooln.core.doc import Doc

from mjooln.atom.zulu import Zulu
from mjooln.atom.key import Key
from mjooln.atom.identity import Identity

logger = logging.getLogger(__name__)


class AtomError(Exception):
    pass


class Atom(Doc):
    """ Unique identifier intended for objects and data sets alike

    The timestamp (zulu) defines start time (t0) of data set, or the time of
    creation. Key defines the grouping, while identity is a unique UUID

    Format: ``<zulu>___<key>___<identity>``

    :class:`.Zulu` represents t0 or creation time

    :class:`.Key` defines grouping of the contents

    :class:`.Identity` is a unique identifier for the contents

    Constructor initializes a valid atom, and will throw an exception
    if a valid atom cannot be created based on input parameters.

    The constructor must as minimum have ``key`` as input::

        a = Atom(key='some_key')
        a.key
            'some_key'
        a.zulu
            Zulu(2020, 5, 22, 13, 13, 18, 179169, tzinfo=<UTC>)
        a.identity
            '060AFBD5_D865_4974_8E37_FDD5C55E7CD8'
        str(a)
            '20200522T131318u179169Z___some_key___060AFBD5_D865_4974_8E37_FDD5C55E7CD8'

    """

    @classmethod
    def is_stub(cls,
                stub: str):
        """
        Check if string is a valid Atom stub

        :param stub: Possible Atom stub
        :return: True if stub is valid, False if not
        :rtype: bool
        """
        parts = stub.split(Name.CLASS_SEPARATOR)
        if len(parts) != 3:
            return False
        z, k, i = parts
        if not Zulu.is_stub(z) \
                or not Key.is_stub(k) \
                or not Identity.is_stub(i):
            return False
        return True

    @classmethod
    def from_stub(cls,
                  stub: str):
        if not cls.is_stub(stub):
            raise AtomError(f'Invalid atom stub: {stub}')
        zulu, key, identity = stub.split(Name.CLASS_SEPARATOR)
        return cls(key=Key(key),
                   zulu=Zulu.from_stub(zulu),
                   identity=Identity(identity))

    @classmethod
    def elf(cls, *args, **kwargs):
        if len(args) == 1 and kwargs is None:
            arg = args[0]
            if isinstance(arg, Atom):
                return arg
            if isinstance(arg, Key):
                return cls(arg)
            elif isinstance(arg, str):
                if Key.is_stub(arg):
                    return cls(arg)
                elif cls.is_stub(arg):
                    return cls.from_stub()
                else:
                    raise AtomError(f'This input string is nowhere near '
                                    f'what I need to create an Atom: {arg}')
            elif Key.is_stub(arg):
                return cls(arg)
            else:
                raise AtomError(f'How the fuck am I supposed to create an atom '
                                f'based on this ridiculous excuse for an '
                                f'input: {arg} [{type(arg)}]')
        elif len(args) == 0:
            if 'key' not in kwargs:
                raise AtomError(f'At the very least, give me a key to work '
                                f'on. You know, key as thoroughly described '
                                f'in class Key')
            key = Key.elf(kwargs['key'])
            identity = None
            zulu = None
            if 'identity' in kwargs:
                identity = Identity.elf(kwargs['identity'])
            if 'zulu' in kwargs:
                zulu = Zulu.elf(kwargs['zulu'])
            return cls(key,
                       zulu=zulu,
                       identity=identity)
        raise AtomError(f'This is rubbish. Cannot make any sense of this '
                        f'mindless junk of input: '
                        f'args={args}; kwargs={kwargs}')

    @classmethod
    def default(cls):
        raise AtomError('Atom does not have a default value')

    def __init__(self,
                 key,
                 zulu: Zulu = None,
                 identity: Identity = None):
        """ Initializes a valid atom

        :param args: One argument is treated as a atom string and parsed.
        Three arguments is treated as (zulu, key, identity), in that order.
        :param kwargs: key, zulu, identity. key is required, while zulu and
        identity are created if not supplied.
        """
        # Initialize Doc class
        super(Atom, self).__init__()
        if isinstance(key, str):
            if Key.is_stub(key):
                key = Key(key)
            elif Atom.is_stub(key):
                raise AtomError(f'Cannot instantiate with Atom stub. '
                                f'Use Atom.from_stub()')
            else:
                raise AtomError(f'Invalid key: {key} [{type(key).__name__}]')

        if not isinstance(key, Key):
            raise AtomError(f'Invalid instance of key: {type(key)}')

        if not zulu:
            zulu = Zulu()
        if not identity:
            identity = Identity()

        self.zulu = zulu
        self.key = key
        self.identity = identity

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return Name.CLASS_SEPARATOR.join([str(self.zulu),
                                          str(self.key),
                                          str(self.identity)])

    def __repr__(self):
        return f'Atom(key=\'{self.key}\', ' \
               f'zulu={self.zulu.__repr__()}, ' \
               f'identity={self.identity.__repr__()})'

    def stub(self):
        return str(self)

    def to_doc(self,
               ignore_private: bool = True):
        return {
            'zulu': self.zulu.iso(),
            'key': str(self.key),
            'identity': str(self.identity),
        }

    @classmethod
    def from_doc(cls, doc: dict):
        return cls(
            zulu=Zulu.from_iso(doc['zulu']),
            key=Key(doc['key']),
            identity=Identity(doc['identity']),
        )

    def with_separator(self,
                       separator: str):
        """ Atom string with custom separator

        :param separator: Custom separator
        :return: str
        """
        return separator.join(str(self).split(Name.CLASS_SEPARATOR))

    @classmethod
    def level_count(self,
                    level: int = None):
        if level is None or level < 0:
            return 1
        else:
            return level

    def levels(self,
               key_level: int = None,
               date_level: int = None,
               time_level: int = 0):
        """ Create list of levels representing the Atom

        Intended usage is for creating folder paths for atom files

        **None**: Full value

        **0**: No value

        **Negative**: Number of elements as string

        **Positive**: Number of elements as list

        Example::

            a = Atom(key='this__is__a_key')
            str(a)
                '20200522T181551u626184Z___this__is__a_key___F45A9C74_DC5D_48A6_9468_FC7E343EC554'
            a.levels()
                ['this__is__a_key', '20200522']
            a.levels(None, None, None)
                ['this__is__a_key', '20200522', '181551']
            a.levels(0, 0, 0)
                []
            a.levels(3, 2, 1)
                ['this', 'is', 'a_key', '2020', '05', '18']
            a.levels(-3, -2, -1)
                ['this__is__a_key', '202005', '18']

        :param key_level: Key levels
        :param date_level: Date levels
        :param time_level: Time levels
        :return: list
        """

        # TODO: Use this to make  level counter as well
        # TODO: Make this as classmethod
        def make(parts, level, sep=''):
            if level == 0:
                return []
            elif level > 0:
                return parts[:level]
            else:
                return [sep.join(parts[:-level])]

        if key_level is None:
            keys = [str(self.key)]
        else:
            # TODO: Use key sep. At least as default
            keys = make(self.key.elements(), key_level, sep=Name.ELEMENT_SEPARATOR)

        zs = self.zulu.str
        if date_level is None:
            dates = [zs.date]
        else:
            dates = make([zs.year, zs.month, zs.day], date_level)
        if time_level is None:
            times = [zs.time]
        else:
            times = make([zs.hour, zs.minute, zs.second], time_level)

        return keys + dates + times



import shutil
import zipfile
import gzip

from mjooln.filesystem.file import File


class ArchiveError(Exception):
    pass


class Archive:
    # TODO: Add gz to zip
    # TODO: Add handling of multiple files and folders in archive

    @classmethod
    def is_zip(cls,
               file: File):
        extensions = file.extensions()
        return len(extensions) > 0 and extensions[-1] == 'zip'

    @classmethod
    def zip_to_gz(cls,
                  file: File,
                  delete_source_file: bool = True):

        if not cls.is_zip(file):
            raise ArchiveError(f'File is not zip-file: {file}')

        with zipfile.ZipFile(file, 'r') as zr:
            file_info = zr.filelist
            if len(file_info) > 1:
                raise ArchiveError(f'Multiple files in archive: {file}')
            elif len(file_info) == 0:
                raise ArchiveError(f'No files in archive: {file}')
            file_info = file_info[0]
            file_name = file_info.filename
            if '/' in file_name:
                file_name = file_name.split('/')[-1]
            if '\\' in file_name:
                file_name = file_name.split('\\')[-1]
            gz_file = File.join(file.folder(), file_name + '.gz')
            with zr.open(file_info, 'r') as zf:
                with gzip.open(gz_file, 'w') as gf:
                    shutil.copyfileobj(zf, gf)

        if delete_source_file:
            file.delete()

        return gz_file





# from .crypt import Crypt, CryptError
# from .dic_doc import Dic, Doc, DocError, JSON
# from .path import Path, PathError
# from .folder import Folder, FolderError
# from .file import File, FileError


import logging
import string
import re

from mjooln.core.name import Name
from mjooln.atom.element import Element, ElementError

logger = logging.getLogger(__name__)


class Key(str):
    """
    Defines key string with limitations

    - Minimum length is 3
    - Allowed characters are
        - Lower case ascii (a-z)
        - Digits (0-9)
        - Underscore (_)
        - Double underscore (__)
    - Underscore and digits can not be the first character
    - Underscore can not be the last character
    - The double underscore act as separator for groups in the key
    - Triple underscore is reserved for separating keys from other keys or
      strings, such as in class :class:`.Atom`

    Sample keys::

        'simple'
        'with_longer_name'
        'group_one__group_two__group_three'
        'digit1'
        'longer_digit2'
        'group1__group2__group3'
        'group_1__group_2__group_3'

    """

    ALLOWED_CHARACTERS = string.ascii_lowercase + string.digits + '_'
    ALLOWED_STARTSWITH = string.ascii_lowercase
    ALLOWED_ENDSWITH = string.ascii_lowercase + string.digits
    MINIMUM_ALLOWED_LENGTH = 3

    REGEX_STRING = r'[a-z][a-z_]*[a-z]'
    RE = re.compile(REGEX_STRING)

    # #: Separates key from other keys or elements, such as identity and zulu
    # #: in Atom
    # OUTER_SEPARATOR = '___'
    #
    # #: Separates groups in key
    # SEPARATOR = '__'

    def __new__(cls,
                key: str):
        # TODO: Add list as input, creating key with separator
        cls.verify_key(key)
        instance = super(Key, cls).__new__(cls, key)
        return instance

    def __repr__(self):
        return f'Key(\'{self}\')'

    def stub(self):
        return self.__str__()

    @classmethod
    def verify_key(cls, key: str):
        if not len(key) >= cls.MINIMUM_ALLOWED_LENGTH:
            raise InvalidKey(f'Key too short. Key \'{key}\' has length '
                             f'{len(key)}, while minimum length is '
                             f'{cls.MINIMUM_ALLOWED_LENGTH}')
        if Name.CLASS_SEPARATOR in key:
            raise InvalidKey(f'Key contains element reserved as class '
                             f'separator. '
                             f'Key \'{key}\' cannot contain '
                             f'\'{Name.CLASS_SEPARATOR}\'')
        elements = key.split(Name.ELEMENT_SEPARATOR)
        for element in elements:
            try:
                Element.verify_element(element)
            except ElementError as ee:
                raise InvalidKey(f'Error with element \'{element}\': {ee}') \
                    from ee

    @classmethod
    def is_stub(cls,
                key: str):
        try:
            cls.verify_key(key)
            return True
        except InvalidKey:
            return False

    def elements(self):
        """ Returns key parts as defined by separator (double underscore)

        Example::

            key = Key('some_key__with_two__no_three_elements')
            key.elements()
                ['some_key', 'with_two', 'three_elements']

        :returns: [str]
        """
        return [Element(x) for x in self.split(Name.ELEMENT_SEPARATOR)]

    def with_separator(self,
                       separator: str):
        """ Replace separator

        Example::

            key = Key('some__key_that_could_be__path')
            key.with_separator('/')
                'some/key_that_could_be/path'

        :param separator: Separator of choice
        :type separator: str
        :return: str
        """
        return separator.join(self.elements())

    @classmethod
    def elf(cls, key):
        """ Allows key class to pass through instead of throwing exception

        :param key: Input key string or key class
        :type key: str or Key
        :return: Key
        """
        if isinstance(key, Key):
            return key
        else:
            return cls(key)


class InvalidKey(Exception):
    pass
